# -*- shell-script -*-
#
# bash-ini: functions for handling ini files in bash
# Copyright (C) 2015-2018 Heinz Deinhart
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# bash-ini version 0.3.1

# _bi_validate_dbname <caller-name> <db-name>
function _bi_validate_dbname() {
    (( $# == 2 )) || \
	bb_fatal "_bi_validate_dbname called with $# args instead of 2"

    local callername="$1"
    local dbname="$2"
    
    [[ "$dbname" =~ ^[[:alnum:]_]+$ ]] || {
	bb_fatal "_bi_validate_dbname: illegal dbname '$dbname'" \
	    "used in: $callername"
    }
}

# _bi_validate_retvn <caller-name> <retvn>
function _bi_validate_retvn() {
    (( $# == 2 )) || \
	bb_fatal "_bi_validate_retvn called with $# args instead of 2"

    local callername="$1"
    local retvn="$2"

    [[ "$_retvn" =~ ^[[:alpha:]][[:alnum:]_]*$ ]] || {
	bb_fatal "_bi_validate_retvn: illegal retvn '$retvn'" \
	    "used in $callername: $dbname"
    }
}

# bi_db_load <db-name> <ini-file>
# loads contents of ini-file
# db-name must match ^[[:alnum:]_]+$ and must be currently unset
# errors: 1=file not readable, 2=variable not unset, 3=parse error
function bi_db_load() {
    local dbname="$1"
    local inputfile="$2"

    _bi_validate_dbname bi_db_load "$dbname"

    [ -r "$inputfile" ] || {
	bb_log_debug 1 "bi_db_load: can not read file: $inputfile"
	return 1
    }

    if bi_db_is_loaded "$dbname"
    then
	bb_log_debug 1 "bi_db_load: abort: db named $dbname is already loaded"
	return 2
    fi

    declare -g -A "bi_db_$dbname" || bb_fatal "bi_db_load: declare failed"


    local -i linenr=0
    local section=

    local extglobsave=$(shopt -p extglob || true)
    shopt -s extglob

    while read -r line || [[ -n "$line" ]]
    do
	linenr=$((linenr + 1))

        # ignore empty lines and comments
	[[ "$line" =~ ^([[:space:]]*|[\;\#].*)$ ]] && {
	    continue
	}

	# sections
	[[ "$line" =~ ^\[([[:alnum:]_\-]([[:alnum:]\ _\-]*[[:alnum:]_\-])?)\][[:space:]]*$ ]] && {
	    section=${BASH_REMATCH[1]}
	    eval "bi_db_$dbname[\"\$section§\"]=''"
	    continue
	}

	# values
	[[ "$line" =~ ^[[:space:]]*([[:alnum:]]([[:alnum:]\ _\-]*[[:alnum:]_\-])?)[[:blank:]]*=[[:blank:]]*(.*)[[:blank:]]*$ ]] && {
	    local key=${BASH_REMATCH[1]}
	    # shellcheck disable=SC2034
	    local value=${BASH_REMATCH[3]}
	    eval "bi_db_$dbname[\"\$section§$key\"]=\"\$value\""
	    continue
	}

	bb_log_debug 1 "bi_db_load: parse error at line $linenr near \"$line\""
	eval "$extglobsave"
	return 3
    done < "$inputfile"

    eval "$extglobsave"
}

# bi_db_destroy <db-name>
# destroys loaded ini db
function bi_db_destroy() {
    local dbname="$1"
    unset "bi_db_$dbname"
}

# bi_db_is_loaded <db-name>
# check if db is loaded
function bi_db_is_loaded() {
    local dbname=$1

    [[ -v "bi_db_${dbname}" ]]
}

# bi_key_exists <db-name> <key> [<section>]
# check if a key exists
# if no section is specified only global context is searched
function bi_key_exists() {
    (( $# == 2 || $# == 3 )) || \
	bb_fatal "bi_key_exists needs 2 or 3 args but called with $#"

    local dbname="$1"
    local key="$2"
    local section="${3-}"

    eval "[[ \"\${bi_db_${dbname}[\"\$section§\$key\"]+isset}\" == \"isset\" ]]"
}

# bi_key_exists_with_default <db-name> <key> <section>
# like bi_key_exists but if section does not exist or contain key it will
# search in global context
function bi_key_exists_with_default() {
    (( $# == 3 )) || \
	bb_fatal "bi_key_exists_with_default needs 3 args but called with $#"

    local dbname="$1"
    local key="$2"
    local section="$3"

    bi_key_exists "$dbname" "$key" "$section" || \
	bi_key_exists "$dbname" "$key"
}

# bi_section_exists <db-name> <section>
# check if a section exists
function bi_section_exists() {
    (( $# == 2 )) || \
	bb_fatal "bi_section_exists needs 2 args but called with $#"

    local dbname="$1"
    local section="$2"

    eval "[[ \"\${bi_db_${dbname}[\"\$section§\"]+isset}\" == \"isset\" ]]"
}

# bi_get_section_list <retval-name> <db-name>
# return a list of all sections as array into global array <retval-name>
# retval-name must match ^[[:alpha:]][[:alnum:]_]*$
# note: the order is undefined
function bi_get_section_list() {
    (( $# == 2 ))  || \
	bb_fatal "bi_get_section_list needs 2 args but called with $#"

    local _retvn="$1"
    local _dbname="$2"

    _bi_validate_retvn "bi_get_section_list" "$_retvn"

    local -a _keys
    eval "_keys=(\"\${!bi_db_${_dbname}[@]}\")"
    local -a _sections=()

    if (( ${#_keys[*]} > 0 )) # skip if file is empty
    then
	local _k
	for _k in "${_keys[@]}"
	do
	    [[ "$_k" =~ §$ ]] && _sections+=("${_k%§}")
	done
    fi

    if [[ "$_retvn" =~ ^[[:alnum:]_]+$ ]]
    then
	if (( ${#_sections[@]} > 0 )) # any sections found?
	then # yes
	    eval "${_retvn}=(\"\${_sections[@]}\")"
	else # no
	    eval "${_retvn}=()"
	fi
    else
	fatal "bi_get_section_list: illegal retval-name: $_retvn"
    fi
}

# bi_get_value <retval-name> <db-name> <key> [ <section> ]
# return value for <key> in <db-name> in optional <section>
# note: if <section> is specified key must exist in it
# note: <retval-name> must not start with '_'
function bi_get_value() {
    (( $# == 3 || $# == 4 )) || \
	bb_fatal "bi_get_value needs 3 or 4 args but called with $#"

    local _retvn="$1"
    local _dbname="$2"
    local _key="$3"
    local _section="${4-}"

    _bi_validate_retvn "bi_get_value" "$_retvn"

    bi_db_is_loaded "$_dbname" || return 1

    bi_key_exists "$_dbname" "$_key" "$_section" || return 1

    eval "${_retvn}=\"\${bi_db_$_dbname[\"\$_section§\$_key\"]}\""
}

# bi_get_value_with_default <variablename> <db-name> <_key> <_section>
# return value for <key> in <db-name> in <section> or use default
# note: <retval-name> must not start with '_'
function bi_get_value_with_default() {
    (( $# == 4 )) || \
	bb_fatal "bi_get_value_with_default needs 4 args but called with $#"

    local _retvn="$1"
    local _dbname="$2"
    local _key="$3"
    local _section="$4"

    _bi_validate_retvn "bi_get_value_with_default" "$_retvn"

    bi_get_value "$_retvn" "$_dbname" "$_key" "$_section" || \
	bi_get_value "$_retvn" "$_dbname" "$_key"
}
